package net.sf.javaprinciples.resource;

import java.io.UnsupportedEncodingException;
import java.net.URLEncoder;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Default implementation of ResourceIdentifier.
 *
 * @author rvllewln
 *
 */
public class DefaultResourceIdentifier implements ResourceIdentifier
{
    private static final int INITIAL_HASH = 17;
    private static final int MAGIC_NUMBER = 31;
    private final List<String> pathElements;
    private final Map<String, String> queryParameters;
    private final String identifier;

    /**
     * Create a ResourceIdentifier.
     *
     * @param pathElements The path elements of the identifier.
     * @param queryParameters The query parameters of the identifier.
     */
    public DefaultResourceIdentifier(String identifier, List<String> pathElements, Map<String, String> queryParameters)
    {
        this.identifier = identifier;
        this.pathElements = pathElements != null ? pathElements : new ArrayList<String>();
        this.queryParameters = queryParameters != null ? queryParameters : new HashMap<String, String>();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public String getIdentifier()
    {
        return identifier;
    }

    @Override
    public List<String> getPathElements()
    {
        return Collections.unmodifiableList(pathElements);
    }

    @Override
    public Map<String, String> getQueryParameters()
    {
        return Collections.unmodifiableMap(queryParameters);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public String toString()
    {
        StringBuffer buffer = new StringBuffer();
        for (String element : getPathElements())
        {
            buffer.append("/").append(element);
        }
        StringBuffer queryBuffer = new StringBuffer();
        boolean isFirst = true;
        for (Map.Entry<String, String> entry : getQueryParameters().entrySet())
        {
            if (isFirst)
            {
                queryBuffer.append("?");
                isFirst = false;
            }
            else
            {
                queryBuffer.append("&");
            }
            queryBuffer.append(entry.getKey()).append("=").append(entry.getValue());
        }
        try
        {
            String url = URLEncoder.encode(queryBuffer.toString(), "UTF-8");
            buffer.append(url);
            return buffer.toString();
        }
        catch (UnsupportedEncodingException uee)
        {
            return buffer.toString();
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public boolean equals(Object other)
    {
        if (!(other instanceof DefaultResourceIdentifier))
        {
            return false;
        }
        DefaultResourceIdentifier toCompare = (DefaultResourceIdentifier)other;
        return toCompare.getPathElements().equals(this.getPathElements()) &&
            toCompare.getQueryParameters().equals(this.getQueryParameters());
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public int hashCode()
    {
        // This sort of thing can be extracted into a utility class.
        int code = MAGIC_NUMBER * INITIAL_HASH + this.getPathElements().hashCode();
        code = code * MAGIC_NUMBER + this.getQueryParameters().hashCode();
        return code;
    }

}
