package net.sf.javaprinciples.resource.servlet;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.springframework.beans.BeanUtils;
import org.springframework.stereotype.Controller;
import org.springframework.ui.Model;
import org.springframework.util.StringUtils;
import org.springframework.web.bind.ServletRequestDataBinder;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestMethod;

import net.sf.javaprinciples.resource.MemoryResource;
import net.sf.javaprinciples.resource.Resource;
import net.sf.javaprinciples.resource.ResourceException;
import net.sf.javaprinciples.resource.ResourceFactory;
import net.sf.javaprinciples.resource.ResourceIdentifier;
import net.sf.javaprinciples.resource.ResourceIdentifierParser;
import net.sf.javaprinciples.resource.ResourceNotFoundException;
import net.sf.javaprinciples.resource.ResourceParseException;

@Controller
public final class ResourceController<T>
{
    private static final Log LOGGER = LogFactory.getLog(ResourceController.class);
    private static final int BUFFER_LENGTH = 2048; // 2k characters
    private ResourceFactory<T> resourceFactory;
    private ResourceIdentifierParser identifierParser;
    private Class<?> commandClass;

    private RequestHandler<T> putHandler = new RequestHandler<T>()
    {
      @Override
      public void handleRequest(ResourceModel<T> model, HttpServletRequest request) throws ResourceException
      {
          ResourceIdentifier identifier = identifierParser.parse(request);
          Resource<T> resource = getResourceFromRequest(request);
          resourceFactory.putResource(identifier, resource);
          // We do this to bypass the response rendering step in StreamView.
          // Arguably we should modify StreamView to detect a missing Resource & assume a PUT.
          model.setStatusCode(HttpServletResponse.SC_OK);
      }
   };
   private RequestHandler<T> getHandler = new RequestHandler<T>()
   {
       @Override
       public void handleRequest(ResourceModel<T> model, HttpServletRequest request) throws ResourceException
       {
           ResourceIdentifier identifier = identifierParser.parse(request);
           Resource<T> resource = resourceFactory.getResource(identifier);
           if (resource != null)
           {
               model.setResource(resource);
           }
           else
           {
               model.setStatusCode(HttpServletResponse.SC_NOT_FOUND);
           }
       }
    };
    private RequestHandler<T> postHandler = new RequestHandler<T>()
    {
        @Override
        @SuppressWarnings("unchecked")
        public void handleRequest(ResourceModel<T> model, HttpServletRequest request) throws ResourceException
        {
            ResourceIdentifier identifier = identifierParser.parse(request);
            Resource<T> resource = getResourceFromRequest(request);
            identifier = resourceFactory.putResource(identifier, resource);
            model.setStatusCode(HttpServletResponse.SC_CREATED);
            model.addProperty("Location", identifier.getIdentifier());
        }
    };
    /**
     * Handle a GET request.
     * @param model the model propagated by Spring through MVC infrastructure.
     * @param request Never null
     */
    @RequestMapping(method = {RequestMethod.GET})
    public void handleGetRequest(Model model, HttpServletRequest request)
    {
        handleRequestInternal(model, request, getHandler);
    }

    /**
     * Handle a PUT request.
     * @param model the model propagated by Spring through MVC infrastructure.
     * @param request Never null
     */
    @RequestMapping(method = {RequestMethod.PUT})
    public void handlePutRequest(Model model, HttpServletRequest request)
    {
        handleRequestInternal(model, request, putHandler);
    }

    /**
     * Handle a POST request.
     * @param model the model propagated by Spring through MVC infrastructure.
     * @param request Never null
     */
    @RequestMapping(method = {RequestMethod.POST})
    public void handlePostRequest(Model model, HttpServletRequest request)
    {
        handleRequestInternal(model, request, postHandler);
    }

    private void handleRequestInternal(Model model, HttpServletRequest request, RequestHandler<T> handler)
    {
        ResourceModel<T> resourceModel = prepareModel(model);

        try
        {
            handler.handleRequest(resourceModel, request);
        }
        // Additional exceptions handled depending on error condition.
        catch (ResourceParseException rse)
        {
            resourceModel.setStatusCode(HttpServletResponse.SC_BAD_REQUEST);
        }
        catch (ResourceNotFoundException re)
        {
            resourceModel.setStatusCode(HttpServletResponse.SC_NO_CONTENT);
        }
        catch (ResourceException re)
        {
            resourceModel.setStatusCode(HttpServletResponse.SC_INTERNAL_SERVER_ERROR);
            LOGGER.error(re.getMessage());
        }
    }

    private ResourceModel<T> prepareModel(Model model)
    {
        ResourceModel<T> resourceModel = new ResourceModel<T>();
        model.addAttribute(ResourceModel.class.getName(), resourceModel);
        return resourceModel;
    }

    // At present we assume that all Resource representations that we will handle
    // will be character-based and not binary, except if a commandClass attribute is configured.
    // In that case, we attempt to bind request parameters to the commandClass and use that for
    // the Resource content - this is to support file uploads primarily. If this assumption
    // changes then this method will have to be changed.
    @SuppressWarnings("unchecked")
    private Resource<T> getResourceFromRequest(HttpServletRequest request) throws ResourceException
    {
        MemoryResource<T> resource = new MemoryResource<T>();
        T content = null;
        if (null != commandClass)
        {
            Object command = BeanUtils.instantiateClass(commandClass);
            ServletRequestDataBinder binder = new ServletRequestDataBinder(command);
            binder.bind(request);
            content = (T)command;
        }
        else
        {
            try
            {
                InputStream is = request.getInputStream();
                if (null != is)
                {
                    ByteArrayOutputStream resourceBuffer = new ByteArrayOutputStream();
                    byte[] cbuf = new byte[BUFFER_LENGTH];
                    int numRead = readBytes(is, resourceBuffer, cbuf);
                    while (numRead >= 0)
                    {
                        numRead = readBytes(is, resourceBuffer, cbuf);
                    }
                    String encoding = request.getCharacterEncoding();
                    if (StringUtils.hasText(encoding))
                    {
                        content = (T)resourceBuffer.toString(encoding);
                        resource.setEncoding(encoding);
                    }
                    else
                    {
                        content = (T)resourceBuffer.toString();
                    }
                }
            }
            catch (IOException ioe)
            {
                // This is a tad dodgy, though the argument can be made that
                // being unable to read in the request stream is sort of a parsing exception.
                throw new ResourceParseException(ioe.getMessage());
            }
        }
        resource.setContent(content);
        resource.setMimeType(getMimeTypeOnly(request));
        String timeToLive = request.getHeader("X-Ttl");
        if (StringUtils.hasText(timeToLive))
        {
            try
            {
                // X-Ttl is in seconds however the lifespan is in milliseconds
                long timeToLiveValue = Long.parseLong(timeToLive) * 1000;
                resource.setLifespan(timeToLiveValue);
            }
            catch (NumberFormatException e)
            {
                // Silently ignore
            }
        }
        return resource;
    }

    protected String getMimeTypeOnly(HttpServletRequest request)
    {
        String contentType = request.getContentType();
        if (!StringUtils.hasText(contentType))
        {
            return null;
        }
        int extra = contentType.indexOf(';');
        if (extra == -1)
        {
            return contentType;
        }
        return contentType.substring(0, extra);
    }

    private int readBytes(InputStream is, ByteArrayOutputStream resourceBuffer, byte[] cbuf) throws IOException
    {
        int numRead = is.read(cbuf);
        if (numRead >= 0)
        {
            resourceBuffer.write(cbuf, 0 ,numRead);
        }
        return numRead;
    }

    private interface RequestHandler<T>
    {
        void handleRequest(ResourceModel<T> model, HttpServletRequest request) throws ResourceException;
    }

    public void setResourceFactory(ResourceFactory<T> resourceFactory)
    {
        this.resourceFactory = resourceFactory;
    }

    public ResourceFactory<T> getResourceFactory()
    {
        return resourceFactory;
    }

    public void setIdentifierParser(ResourceIdentifierParser identifierParser)
    {
        this.identifierParser = identifierParser;
    }

    public void setCommandClass(Class<?> commandClass)
    {
        this.commandClass = commandClass;
    }
}
